@@ -23,12 +23,10 @@
 #include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
+#include <string.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#ifdef HAVE_STRING_H
-#include <string.h>
-#endif
 #ifdef HAVE_POSTGRESQL_LIBPQ_FE_H
 #include <postgresql/libpq-fe.h>
 #else
@@ -50,6 +48,75 @@
 
 #define PROB(x)  (0 > (x))
 
+
+/* Abstractions for Scheme objects to C string conversion.  */
+
+typedef struct {
+  char *s;
+  size_t len;
+} range_t;
+
+#define RS(svar)    c ## svar .s
+#define RLEN(svar)  c ## svar .len
+
+#if GI_LEVEL_NOT_YET_1_8
+
+#define FINANGLABLE_SCHEME_STRING_FROM_SYMBOL(sym)      \
+  scm_string_copy (scm_symbol_to_string (sym))
+
+/* Coerce a Scheme (sub)string that is to be used in contexts where the
+   extracted C string is expected to be read-only and zero-terminated.  We
+   check this condition precisely instead of simply coercing all (sub)strings,
+   to avoid waste for those substrings that may in fact (by lucky accident)
+   already satisfy the condition.  */
+#define ROZT_X(x)                                       \
+  if (SCM_ROCHARS (x) [SCM_ROLENGTH (x)])               \
+    x = BSTRING (SCM_ROCHARS (x), SCM_ROLENGTH (x))
+
+#define _FINANGLE(svar,p1)  do                  \
+    {                                           \
+      if (p1)                                   \
+        ROZT_X (svar);                          \
+      RS (svar) = SCM_ROCHARS (svar);           \
+      RLEN (svar) = SCM_ROLENGTH (svar);        \
+    }                                           \
+  while (0)
+
+#define UNFINANGLE(svar)
+
+#else  /* !GI_LEVEL_NOT_YET_1_8 */
+
+#define FINANGLABLE_SCHEME_STRING_FROM_SYMBOL  scm_symbol_to_string
+
+#define REND(svar)          RS (svar) [RLEN (svar)]
+#define NUL_AT_END_X(svar)  REND (svar) = '\0'
+
+#define _FINANGLE(svar,p1)  do                                  \
+    {                                                           \
+      RS (svar) = scm_to_locale_stringn (svar, &RLEN (svar));   \
+      if (RS (svar))                                            \
+        {                                                       \
+          if (p1 && REND (svar))                                \
+            {                                                   \
+              RS (svar) = realloc (RS (svar), 1 + RLEN (svar)); \
+              NUL_AT_END_X (svar);                              \
+            }                                                   \
+        }                                                       \
+      else                                                      \
+        RS (svar) = strdup ("");                                \
+    }                                                           \
+  while (0)
+
+#define UNFINANGLE(svar)  free (RS (svar))
+
+#endif  /* !GI_LEVEL_NOT_YET_1_8 */
+
+/* Use ‘FINANGLE_RAW’ when the consumer of the C string takes full range
+   (start address plus length) info.  Otherwise, ‘FINANGLE’.  */
+
+#define FINANGLE_RAW(svar)  _FINANGLE (svar, 0)
+#define FINANGLE(svar)      _FINANGLE (svar, 1)
+
 
 /*
  * smob: connection
@@ -127,10 +194,10 @@ xc_free (SCM obj)
 
   if (xc->dbconn)
     PQfinish (xc->dbconn);
-  free (xc);
+  GCFREE (xc, xc_name);
   SCM_SET_SMOB_DATA (obj, NULL);
 
-  return sizeof (xc_t);
+  return GCRV (xc);
 }
 
 #define CONN_NOTICE(conn)   (xc_unbox (conn)->notice)
@@ -170,6 +237,29 @@ typedef struct
 #define OPLOBPORTP(x)    (LOBPORT_WITH_FLAGS_P (x, SCM_OPN))
 #define OPINLOBPORTP(x)  (LOBPORT_WITH_FLAGS_P (x, SCM_OPN | SCM_RDNG))
 
+static long
+extract_mode_bits (SCM modes, int *appendp)
+{
+  range_t cmodes;
+  long rv;
+
+  FINANGLE (modes);
+  rv = scm_mode_bits (RS (modes));
+  if (appendp)
+    *appendp = strchr (RS (modes), 'a') ? 1 : 0;
+  UNFINANGLE (modes);
+  return rv;
+}
+
+#define ASSERT_MODES_COPY(pos,svar,cvar,ap)  do \
+    {                                           \
+      ASSERT_STRING (pos, svar);                \
+      cvar = extract_mode_bits (svar, ap);      \
+    }                                           \
+  while (0)
+
+static char lob_name[] = "PG-LO-PORT";
+
 static SCM
 lob_mklobport (SCM conn, Oid oid, int alod, long modes, const char *FUNC_NAME)
 {
@@ -177,9 +267,9 @@ lob_mklobport (SCM conn, Oid oid, int alod, long modes, const char *FUNC_NAME)
   lob_stream *lobp;
   scm_port *pt;
 
-  lobp = (lob_stream *) scm_must_malloc (sizeof (lob_stream), "PG-LO-PORT");
+  lobp = GCMALLOC (sizeof (lob_stream), lob_name);
 
-  SCM_NEWCELL (port);
+  NEWCELL_X (port);
 
   NOINTS ();
   lobp->conn = conn;
@@ -252,10 +342,7 @@ throws a @code{misc-error} if the @code{modes} string is invalid.  */)
   int pg_modes = 0;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
-  ASSERT_STRING (2, modes);
-  ROZT_X (modes);
-
-  mode_bits = scm_mode_bits (ROZT (modes));
+  ASSERT_MODES_COPY (2, modes, mode_bits, NULL);
 
   if (mode_bits & SCM_RDNG)
     pg_modes |= INV_READ;
@@ -307,7 +394,7 @@ some idea of what happened.
 Throw @code{misc-error} if the @code{modes} is invalid.  */)
 {
 #define FUNC_NAME s_pg_lo_open
-  long mode_bits;
+  long mode_bits; int appendp;
   PGconn *dbconn;
   int alod;
   Oid pg_oid;
@@ -315,10 +402,7 @@ Throw @code{misc-error} if the @code{modes} is invalid.  */)
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   SCM_VALIDATE_ULONG_COPY (2, oid, pg_oid);
-  ASSERT_STRING (3, modes);
-  ROZT_X (modes);
-
-  mode_bits = scm_mode_bits (ROZT (modes));
+  ASSERT_MODES_COPY (3, modes, mode_bits, &appendp);
 
   if (mode_bits & SCM_RDNG)
     pg_modes |= INV_READ;
@@ -334,7 +418,7 @@ Throw @code{misc-error} if the @code{modes} is invalid.  */)
   if (PROB (alod))
     RETURN_FALSE ();
 
-  if (strchr (ROZT (modes), 'a'))
+  if (appendp)
     {
       NOINTS ();
       if (PROB (lo_lseek (dbconn, alod, 0, SEEK_END)))
@@ -552,14 +636,13 @@ relative to the beginning of the object, or a number less than
 zero if an error occurred.  */)
 {
 #define FUNC_NAME s_pg_lo_seek
-  int cwhere, cwhence;
   ASSERT_PORT (1, port, OPLOBPORTP);
-  SCM_VALIDATE_INUM_COPY (2, where, cwhere);
-  SCM_VALIDATE_INUM_COPY (3, whence, cwhence);
+  ASSERT_EXACT (2, where);
+  ASSERT_EXACT (3, whence);
 
   lob_flush (port);
 
-  return NUM_INT (lob_seek (port, cwhere, cwhence));
+  return NUM_INT (lob_seek (port, C_INT (where), C_INT (whence)));
 #undef FUNC_NAME
 }
 
@@ -649,8 +732,8 @@ Return a string containing the data read from the port or
   char *stage, *wp;
   int csiz, cnum, len, c;
 
-  SCM_VALIDATE_INUM_MIN_COPY (1, siz, 0, csiz);
-  SCM_VALIDATE_INUM_MIN_COPY (2, num, 0, cnum);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (1, siz, csiz);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (2, num, cnum);
   ASSERT_PORT (3, port, OPINLOBPORTP);
 
   if (0 > (len = csiz * cnum)
@@ -700,8 +783,8 @@ lob_free (SCM port)
 
   if (SCM_OPENP (port))
     ret = lob_close (port);
-  free (lobp);
-  return 0;
+  GCFREE (lobp, lob_name);
+  return GCRV (lobp);
 }
 
 PRIMPROC
@@ -718,14 +801,16 @@ the failure.  */)
 {
 #define FUNC_NAME s_pg_lo_import
   PGconn *dbconn;
+  range_t cfilename;
   Oid ret;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, filename);
-  ROZT_X (filename);
 
   NOINTS ();
-  ret = lo_import (dbconn, ROZT (filename));
+  FINANGLE (filename);
+  ret = lo_import (dbconn, RS (filename));
+  UNFINANGLE (filename);
   INTSOK ();
 
   return DEFAULT_FALSE (InvalidOid != ret,
@@ -747,15 +832,17 @@ of the failure.  */)
 #define FUNC_NAME s_pg_lo_export
   PGconn *dbconn;
   Oid pg_oid;
+  range_t cfilename;
   int ret;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   SCM_VALIDATE_ULONG_COPY (2, oid, pg_oid);
   ASSERT_STRING (3, filename);
-  ROZT_X (filename);
 
   NOINTS ();
-  ret = lo_export (dbconn, pg_oid, ROZT (filename));
+  FINANGLE (filename);
+  ret = lo_export (dbconn, pg_oid, RS (filename));
+  UNFINANGLE (filename);
   INTSOK ();
 
   return BOOLEAN (! PROB (ret));
@@ -893,12 +980,11 @@ struct paramspecs
 #define VALIDATE_PARAM_RELATED_ARGS(what)               \
   do {                                                  \
     VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn); \
-    ROZT_X (what);                                      \
+    SCM_VALIDATE_STRING (2, what);                      \
     SCM_VALIDATE_VECTOR (3, parms);                     \
+    FINANGLE (what);                                    \
   } while (0)
 
-#define VREF(v,i)  (SCM_VELTS (v)[i])
-
 static void
 prep_paramspecs (const char *FUNC_NAME, struct paramspecs *ps, SCM v)
 {
@@ -922,7 +1008,14 @@ prep_paramspecs (const char *FUNC_NAME, struct paramspecs *ps, SCM v)
   if (! ps->values)
     MEMORY_ERROR ();
   for (i = 0; i < len; i++)
-    ps->values[i] = ROZT (VREF (v, i));
+    {
+      range_t celem;
+
+      elem = VREF (v, i);
+      FINANGLE (elem);
+      ps->values[i] = strdup (RS (elem));
+      UNFINANGLE (elem);
+    }
   ps->lengths = NULL;
   ps->formats = NULL;
 }
@@ -942,15 +1035,13 @@ drop_paramspecs (struct paramspecs *ps)
  */
 
 #define VALIDATE_FIELD_NUMBER_COPY(pos,num,res,cvar)                    \
-  SCM_VALIDATE_INUM_RANGE_COPY (pos, num, 0, PQnfields (res), cvar)
+  VALIDATE_EXACT_0_UP_TO_N_COPY (pos, num, PQnfields (res), cvar)
 
 
 /*
  * meta and connection
  */
 
-static SCM goodies;
-
 PRIMPROC
 (pg_guile_pg_loaded, "pg-guile-pg-loaded", 0, 0, 0,
  (void),
@@ -966,7 +1057,43 @@ available.  You can test this like so:
 (false-if-exception (pg-guile-pg-loaded))
 @end lisp  */)
 {
-  return goodies;
+  SCM rv = SCM_EOL;
+  /* $ ttn-do generate-C-symbol-set     \
+     PQPROTOCOLVERSION                  \
+     PQRESULTERRORMESSAGE               \
+     PQPASS                             \
+     PQBACKENDPID                       \
+     PQOIDVALUE                         \
+     PQBINARYTUPLES                     \
+     PQFMOD                             \
+     PQSETNONBLOCKING                   \
+     PQISNONBLOCKING                    \
+     -b char  */
+  static const char symbolpool[126] =
+    {
+      9 /* count */,
+      17,'P','Q','P','R','O','T','O','C','O','L','V','E','R','S','I','O','N',
+      20,'P','Q','R','E','S','U','L','T','E','R','R','O','R','M','E','S','S','A','G','E',
+      6,'P','Q','P','A','S','S',
+      12,'P','Q','B','A','C','K','E','N','D','P','I','D',
+      10,'P','Q','O','I','D','V','A','L','U','E',
+      14,'P','Q','B','I','N','A','R','Y','T','U','P','L','E','S',
+      6,'P','Q','F','M','O','D',
+      16,'P','Q','S','E','T','N','O','N','B','L','O','C','K','I','N','G',
+      15,'P','Q','I','S','N','O','N','B','L','O','C','K','I','N','G'
+    };
+  const char *p = symbolpool;
+  int count = *p++;
+
+  while (count--)
+    {
+      int len = *p++;
+      SCM sym = scm_string_to_symbol (BSTRING (p, len));
+
+      rv = CONS (sym, rv);
+      p += len;
+    }
+  return rv;
 }
 
 PRIMPROC
@@ -1157,12 +1284,14 @@ need to be escaped a second time.  */)
 #define FUNC_NAME s_pg_connectdb
   xc_t *xc;
   PGconn *dbconn;
+  range_t cconstr;
 
   ASSERT_STRING (1, constr);
-  ROZT_X (constr);
 
   NOINTS ();
-  dbconn = PQconnectdb (ROZT (constr));
+  FINANGLE (constr);
+  dbconn = PQconnectdb (RS (constr));
+  UNFINANGLE (constr);
 
   if (CONNECTION_BAD == PQstatus (dbconn))
     {
@@ -1175,7 +1304,7 @@ need to be escaped a second time.  */)
     }
   INTSOK ();
 
-  xc = ((xc_t *) scm_must_malloc (sizeof (xc_t), xc_name));
+  xc = GCMALLOC (sizeof (xc_t), xc_name);
 
   xc->dbconn = dbconn;
   xc->notice = SCM_BOOL_T;
@@ -1296,23 +1425,27 @@ If there is an error, return @code{#f}.  */)
 {
 #define FUNC_NAME s_pg_escape_string_conn
   PGconn *dbconn;
+  range_t cstring;
   char *answer;
-  size_t ilen, olen;
+  size_t olen;
   int errcode;
   SCM rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, string);
-  ROZT_X (string);
 
-  ilen = SCM_ROLENGTH (string);
-  if (! (answer = malloc (1 + 2 * ilen)))
-    SYSTEM_ERROR ();
+  FINANGLE_RAW (string);
+  if (! (answer = malloc (1 + 2 * RLEN (string))))
+    {
+      UNFINANGLE (string);
+      SYSTEM_ERROR ();
+    }
 
-  olen = PQescapeStringConn (dbconn, answer, ROZT (string), ilen, &errcode);
+  olen = PQescapeStringConn (dbconn, answer, RS (string), RLEN (string), &errcode);
   rv = DEFAULT_FALSE (! errcode,
                       STRING (answer));
   free (answer);
+  UNFINANGLE (string);
   return rv;
 #undef FUNC_NAME
 }
@@ -1329,18 +1462,21 @@ The returned bytea does not have surrounding single-quote chars.
 If there is an error, return @code{#f}.  */)
 {
 #define FUNC_NAME s_pg_escape_bytea_conn
+  void *ucbytea; range_t cbytea;
   PGconn *dbconn;
   unsigned char *answer;
-  size_t ilen, olen;
+  size_t olen;
   SCM rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, bytea);
 
-  ilen = SCM_ROLENGTH (bytea);
+  FINANGLE_RAW (bytea);
+  ucbytea = RS (bytea);
   rv = DEFAULT_FALSE
-    (answer = PQescapeByteaConn (dbconn, SCM_ROUCHARS (bytea), ilen, &olen),
+    (answer = PQescapeByteaConn (dbconn, ucbytea, RLEN (bytea), &olen),
      BSTRING ((char *) answer, olen ? olen - 1 : 0));
+  UNFINANGLE (bytea);
   PQfreemem (answer);
   return rv;
 #undef FUNC_NAME
@@ -1355,15 +1491,18 @@ If there is an error, return @code{#f}.  */)
 {
 #define FUNC_NAME s_pg_unescape_bytea
   unsigned char *answer;
+  void *ucbytea; range_t cbytea;
   size_t olen;
   SCM rv;
 
   ASSERT_STRING (1, bytea);
-
+  FINANGLE_RAW (bytea);
+  ucbytea = RS (bytea);
   rv = DEFAULT_FALSE
-    (answer = PQunescapeBytea (SCM_ROUCHARS (bytea), &olen),
+    (answer = PQunescapeBytea (ucbytea, &olen),
      BSTRING ((char *) answer, olen));
   PQfreemem (answer);
+  UNFINANGLE (bytea);
   return rv;
 #undef FUNC_NAME
 }
@@ -1393,16 +1532,18 @@ message is available only until the next call to @code{pg-exec}
 on this connection.  */)
 {
 #define FUNC_NAME s_pg_exec
+  range_t cstatement;
   SCM z;
   PGconn *dbconn;
   PGresult *result;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, statement);
-  ROZT_X (statement);
 
   NOINTS ();
-  result = PQexec (dbconn, ROZT (statement));
+  FINANGLE (statement);
+  result = PQexec (dbconn, RS (statement));
+  UNFINANGLE (statement);
 
   z = res_box (result);
   INTSOK ();
@@ -1420,6 +1561,7 @@ parameterized string, and @var{parms} is a parameter-vector.  */)
 #define FUNC_NAME s_pg_exec_params
   SCM z;
   PGconn *dbconn;
+  range_t cstatement;
   PGresult *result;
   struct paramspecs ps;
 
@@ -1427,9 +1569,10 @@ parameterized string, and @var{parms} is a parameter-vector.  */)
 
   prep_paramspecs (FUNC_NAME, &ps, parms);
   NOINTS ();
-  result = PQexecParams (dbconn, ROZT (statement), ps.len,
+  result = PQexecParams (dbconn, RS (statement), ps.len,
                          ps.types, ps.values, ps.lengths, ps.formats,
                          RESFMT_TEXT);
+  UNFINANGLE (statement);
   z = res_box (result);
   INTSOK ();
   drop_paramspecs (&ps);
@@ -1450,6 +1593,7 @@ name specified in some prior SQL @code{PREPARE} statement.
 #define FUNC_NAME s_pg_exec_prepared
   SCM z;
   PGconn *dbconn;
+  range_t cstname;
   PGresult *result;
   struct paramspecs ps;
 
@@ -1457,9 +1601,10 @@ name specified in some prior SQL @code{PREPARE} statement.
 
   prep_paramspecs (FUNC_NAME, &ps, parms);
   NOINTS ();
-  result = PQexecPrepared (dbconn, ROZT (stname), ps.len,
+  result = PQexecPrepared (dbconn, RS (stname), ps.len,
                            ps.values, ps.lengths, ps.formats,
                            RESFMT_TEXT);
+  UNFINANGLE (stname);
   z = res_box (result);
   INTSOK ();
   drop_paramspecs (&ps);
@@ -1525,7 +1670,7 @@ A symbol.
   char *s;
 
   VALIDATE_RESULT_UNBOX (1, result, res);
-  SCM_VALIDATE_KEYWORD (2, fieldcode);
+  VALIDATE_KEYWORD (2, fieldcode);
 
 #define CHKFC(x,y)                              \
   if (EQ (fieldcode, (x)))                 \
@@ -1822,12 +1967,16 @@ or @code{#f} if there is no such parameter.
 {
 #define FUNC_NAME s_pg_parameter_status
   PGconn *dbconn;
+  range_t cparm;
   const char *cstatus = NULL;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   SCM_VALIDATE_SYMBOL (2, parm);
+  parm = FINANGLABLE_SCHEME_STRING_FROM_SYMBOL (parm);
 
-  cstatus = PQparameterStatus (dbconn, ROZT (parm));
+  FINANGLE (parm);
+  cstatus = PQparameterStatus (dbconn, RS (parm));
+  UNFINANGLE (parm);
   return DEFAULT_FALSE (cstatus, STRING (cstatus));
 #undef FUNC_NAME
 }
@@ -1972,14 +2121,16 @@ otherwise.  */)
 {
 #define FUNC_NAME s_pg_fnumber
   PGresult *res;
+  range_t cfname;
   int fnum;
 
   VALIDATE_RESULT_UNBOX (1, result, res);
   ASSERT_STRING (2, fname);
-  ROZT_X (fname);
 
   NOINTS ();
-  fnum = PQfnumber (res, ROZT (fname));
+  FINANGLE (fname);
+  fnum = PQfnumber (res, RS (fname));
+  UNFINANGLE (fname);
   INTSOK ();
 
   return NUM_INT (fnum);
@@ -2108,8 +2259,8 @@ up to the caller to convert this to the required type.  */)
   SCM rv;
 
   VALIDATE_RESULT_UNBOX (1, result, res);
-  SCM_VALIDATE_INUM_MIN_COPY (2, stuple, 0, ctuple);
-  SCM_VALIDATE_INUM_MIN_COPY (3, sfield, 0, cfield);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (2, stuple, ctuple);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (3, sfield, cfield);
   CHECK_TUPLE_COORDS ();
 
   NOINTS ();
@@ -2136,8 +2287,8 @@ Return the size in bytes of the value of the attribute
   SCM ret;
 
   VALIDATE_RESULT_UNBOX (1, result, res);
-  SCM_VALIDATE_INUM_MIN_COPY (2, stuple, 0, ctuple);
-  SCM_VALIDATE_INUM_MIN_COPY (3, sfield, 0, cfield);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (2, stuple, ctuple);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (3, sfield, cfield);
   CHECK_TUPLE_COORDS ();
 
   NOINTS ();
@@ -2162,8 +2313,8 @@ Return @code{#t} if the value of the attribute @var{sfield}, tuple
   SCM rv;
 
   VALIDATE_RESULT_UNBOX (1, result, res);
-  SCM_VALIDATE_INUM_MIN_COPY (2, stuple, 0, ctuple);
-  SCM_VALIDATE_INUM_MIN_COPY (3, sfield, 0, cfield);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (2, stuple, ctuple);
+  ASSERT_EXACT_NON_NEGATIVE_COPY (3, sfield, cfield);
   CHECK_TUPLE_COORDS ();
 
   NOINTS ();
@@ -2224,14 +2375,16 @@ and the attempt would block; and @code{-1} if an error occurred.  */)
 {
 #define FUNC_NAME s_pg_put_copy_data
   PGconn *dbconn;
+  range_t cdata;
+  int rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   SCM_VALIDATE_STRING (2, data);
 
-  ROZT_X (data);
-  return NUM_INT (PQputCopyData (dbconn,
-                                 ROZT (data),
-                                 SCM_ROLENGTH (data)));
+  FINANGLE_RAW (data);
+  rv = PQputCopyData (dbconn, RS (data), RLEN (data));
+  UNFINANGLE (data);
+  return NUM_INT (rv);
 #undef FUNC_NAME
 }
 
@@ -2248,17 +2401,20 @@ and the attempt would block; and @code{-1} if an error occurred.  */)
 {
 #define FUNC_NAME s_pg_put_copy_end
   PGconn *dbconn;
-  char *cerrmsg = NULL;
+  range_t cerrmsg;
+  int rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
+  RS (errmsg) = NULL;
   if (GIVENP (errmsg))
     {
       SCM_VALIDATE_STRING (2, errmsg);
-      ROZT_X (errmsg);
-      cerrmsg = ROZT (errmsg);
+      FINANGLE (errmsg);
     }
 
-  return NUM_INT (PQputCopyEnd (dbconn, cerrmsg));
+  rv = PQputCopyEnd (dbconn, RS (errmsg));
+  UNFINANGLE (errmsg);
+  return NUM_INT (rv);
 #undef FUNC_NAME
 }
 
@@ -2269,10 +2425,16 @@ PRIMPROC
 Get a line of @code{COPY} data from @var{conn}, writing it to
 output @var{port}.  If @var{port} is a pair, construct
 a new string and set its @sc{car} to the new string.
+Optional arg @var{async?} non-@code{#f} means don't block
+if there is no data available yet.
 Return the number of data bytes in the row (always greater
 than zero); zero to mean the @code{COPY} is still in progress
 and no data is yet available; @code{-1} to mean the @code{COPY} is
-done; or @code{-2} to mean an error occurred.  */)
+done; or @code{-2} to mean an error occurred.
+
+-args: (- 1 0 async?)
+
+*/)
 {
 #define FUNC_NAME s_pg_get_copy_data
   PGconn *dbconn;
@@ -2311,7 +2473,10 @@ PRIMPROC
 Read a line from @var{conn} on which a @code{COPY <table> TO
 STDOUT} has been issued.  Return a string from the connection.
 A returned string consisting of a backslash followed by a full
-stop signifies an end-of-copy marker.  */)
+stop signifies an end-of-copy marker.
+
+This procedure is obsolete and @strong{WILL BE REMOVED}
+by 2012-12-31.  */)
 {
 #define FUNC_NAME s_pg_getline
   PGconn *dbconn;
@@ -2344,10 +2509,14 @@ Return @code{-1} to mean end-of-copy marker recognized, or a number
 The returned data may contain at most one newline (in the last
 byte position).
 Optional arg @var{tickle} non-@code{#f} means to do a
-``consume input'' operation prior to the read.  */)
+``consume input'' operation prior to the read.
+
+This procedure is obsolete and @strong{WILL BE REMOVED}
+by 2012-12-31.  */)
 {
 #define FUNC_NAME s_pg_getlineasync
   PGconn *dbconn;
+  int rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, buf);
@@ -2358,9 +2527,25 @@ Optional arg @var{tickle} non-@code{#f} means to do a
        the first place.  */
     PQconsumeInput (dbconn);
 
-  return NUM_INT (PQgetlineAsync (dbconn,
-                                  SCM_ROCHARS (buf),
-                                  SCM_ROLENGTH (buf)));
+#if GI_LEVEL_NOT_YET_1_8
+  rv = PQgetlineAsync (dbconn, SCM_ROCHARS (buf), SCM_ROLENGTH (buf));
+#else  /* !GI_LEVEL_NOT_YET_1_8 */
+  /* Fill a temporary C buffer then copy into the Scheme string.
+     Totally gross!  */
+  {
+    size_t sz = scm_c_string_length (buf);
+    char *cbuf = malloc (sz);
+
+    if (! cbuf)
+      SYSTEM_ERROR ();
+    rv = PQgetlineAsync (dbconn, cbuf, sz);
+    while (sz--)
+      scm_c_string_set_x (buf, sz, CHARACTER (cbuf[sz]));
+    free (cbuf);
+  }
+#endif /* !GI_LEVEL_NOT_YET_1_8 */
+
+  return NUM_INT (rv);
 #undef FUNC_NAME
 }
 
@@ -2379,11 +2564,14 @@ Return @code{#t} if successful.  */)
 #define FUNC_NAME s_pg_putline
   PGconn *dbconn;
   int status;
+  range_t cstr;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, str);
   NOINTS ();
-  status = PQputnbytes (dbconn, SCM_ROCHARS (str), SCM_ROLENGTH (str));
+  FINANGLE_RAW (str);
+  status = PQputnbytes (dbconn, RS (str), RLEN (str));
+  UNFINANGLE (str);
   INTSOK ();
   return BOOLEAN (! status);
 #undef FUNC_NAME
@@ -2428,7 +2616,7 @@ verbosity.  */)
   PGconn *dbconn;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
-  SCM_VALIDATE_KEYWORD (2, verbosity);
+  VALIDATE_KEYWORD (2, verbosity);
 
   {
     PGVerbosity now = PQERRORS_DEFAULT;
@@ -2537,47 +2725,34 @@ sepo_unbox (SCM obj)
   return SMOBDATA (obj);
 }
 
+static char sepo_name[] = "PG-PRINT-OPTION";
+
 static size_t
 sepo_free (SCM obj)
 {
   PQprintOpt *po = sepo_unbox (obj);
-  size_t size = 0;
 
-#define _FREE_STRING(p)                         \
-  do {                                          \
-    if (p)                                      \
-      {                                         \
-        size += 1 + strlen (p);                 \
-        free (p);                               \
-      }                                         \
-  } while (0)
-
-  _FREE_STRING (po->fieldSep);
-  _FREE_STRING (po->tableOpt);
-  _FREE_STRING (po->caption);
+  free (po->fieldSep);
+  free (po->tableOpt);
+  free (po->caption);
 
   if (po->fieldName)
     {
       int i = 0;
       while (po->fieldName[i])
         {
-          _FREE_STRING (po->fieldName[i]);
+          free (po->fieldName[i]);
           i++;
         }
-      size += i * sizeof (char *);
       free (po->fieldName);
     }
-#undef _FREE_STRING
 
-  size += sizeof (PQprintOpt);
-  free (po);
+  GCFREE (po, sepo_name);
   SCM_SET_SMOB_DATA (obj, NULL);
 
-  return size;
+  return GCRV (po);
 }
 
-static char sepo_name[] = "PG-PRINT-OPTION";
-
 static int
 sepo_display (UNUSED SCM sepo, SCM port, UNUSED scm_print_state *pstate)
 {
@@ -2622,6 +2797,28 @@ SCM_SYMBOL (pg_sym_field_names, "field-names");
 static SCM valid_print_option_flags;
 static SCM valid_print_option_keys;
 
+static char *
+option_as_string (SCM alist, SCM key, const char *def)
+{
+  SCM maybe = scm_assq_ref (alist, key);
+
+  if (NOT_FALSEP (maybe))
+    {
+      range_t cmaybe;
+      char *rv;
+
+      FINANGLE (maybe);
+      rv = strdup (RS (maybe));
+      UNFINANGLE (maybe);
+      return rv;
+    }
+
+  if (def)
+    return strdup (def);
+
+  return NULL;
+}
+
 PRIMPROC
 (pg_make_print_options, "pg-make-print-options", 1, 0, 0,
  (SCM spec),
@@ -2632,28 +2829,21 @@ of elements, each either a flag (symbol) or a key-value pair
 (with the key being a symbol).  Recognized flags:
 
 @itemize
-@item header: Print output field headings and row count.
-@item align: Fill align the fields.
-@item standard: Old brain-dead format.
-@item html3: Output HTML tables.
-@item expanded: Expand tables.
+@item @t{header} --- Print output field headings and row count.
+@item @t{align} --- Fill align the fields.
+@item @t{standard} --- Old brain-dead format.
+@item @t{html3} --- Output HTML tables.
+@item @t{expanded} --- Expand tables.
 @end itemize
 
 To specify a disabled flag, use @dfn{no-FLAG}, e.g.,
 @code{no-header}.  Recognized keys:
 
 @itemize
-@item field-sep
-String specifying field separator.
-
-@item table-opt
-String specifying HTML table attributes.
-
-@item caption
-String specifying caption to use in HTML table.
-
-@item field-names
-List of replacement field names, each a string.
+@item @t{field-sep} --- String specifying field separator.
+@item @t{table-opt} --- String specifying HTML table attributes.
+@item @t{caption} --- String specifying caption to use in HTML table.
+@item @t{field-names} --- List of replacement field names, each a string.
 @end itemize  */)
 {
 #define FUNC_NAME s_pg_make_print_options
@@ -2705,7 +2895,7 @@ List of replacement field names, each a string.
 #undef CHECK_HEAD
     }
 
-  po = scm_must_malloc (sizeof (PQprintOpt), sepo_name);
+  po = GCMALLOC (sizeof (PQprintOpt), sepo_name);
 
 #define _FLAG_CHECK(m)                                  \
   (MEMQ (pg_sym_no_ ## m, flags)                        \
@@ -2721,14 +2911,8 @@ List of replacement field names, each a string.
 #undef _FLAG_CHECK
 
 #define _STRING_CHECK_SETX(k,m)                         \
-  do {                                                  \
-    SCM stemp = scm_assq_ref (keys, pg_sym_ ## k);      \
-    po->m = (NOT_FALSEP (stemp)                         \
-             ? strdup (SCM_ROCHARS (stemp))             \
-             : (default_print_options.m                 \
-                ? strdup (default_print_options.m)      \
-                : NULL));                               \
-  } while (0)
+  po->m = option_as_string                              \
+    (keys, pg_sym_ ## k, default_print_options.m)
 
   _STRING_CHECK_SETX (field_sep, fieldSep);
   _STRING_CHECK_SETX (table_opt, tableOpt);
@@ -2740,12 +2924,16 @@ List of replacement field names, each a string.
   else
     {
       int i;
-      po->fieldName = (char **) scm_must_malloc ((1 + count) * sizeof (char *),
-                                                 sepo_name);
+      po->fieldName = GCMALLOC ((1 + count) * sizeof (char *), sepo_name);
       po->fieldName[count] = NULL;
       for (i = 0; i < count; i++)
         {
-          po->fieldName[i] = strdup (SCM_ROCHARS (CAR (substnames)));
+          SCM name = CAR (substnames);
+          range_t cname;
+
+          FINANGLE (name);
+          po->fieldName[i] = strdup (RS (name));
+          UNFINANGLE (name);
           substnames = CDR (substnames);
         }
     }
@@ -2935,39 +3123,48 @@ is out of range.  If @var{start} is exactly the length of
 @var{string}, return zero.  */)
 {
 #define FUNC_NAME s_pg_mblen
-  SCM cell;
-  int cenc;
-  size_t cstart = 0, clen;
+  SCM as_string, cell;
+  range_t cstring;
+  int cenc, rv;
+  size_t cstart = 0;
 
   if (STRINGP (encoding))
-    encoding = SYMBOL (ROZT (encoding));
+    encoding = scm_string_to_symbol (encoding);
   ASSERT (encoding, SYMBOLP (encoding), 1);
+  as_string = FINANGLABLE_SCHEME_STRING_FROM_SYMBOL (encoding);
   cell = scm_assq (encoding, CDR (encoding_alist));
   if (NOT_FALSEP (cell))
     cenc = C_INT (CDR (cell));
   else
     {
-      if (PROB (cenc = pg_char_to_encoding (ROZT (encoding))))
+      range_t cas_string;
+
+      FINANGLE (as_string);
+      cenc = pg_char_to_encoding (RS (as_string));
+      UNFINANGLE (as_string);
+      if (PROB (cenc))
         ERROR ("No such encoding: ~A", encoding);
       cell = CONS (encoding, NUM_INT (cenc));
       SETCDR (encoding_alist, CONS (cell, CDR (encoding_alist)));
     }
 
   ASSERT_STRING (2, string);
-  clen = SCM_ROLENGTH (string);
+  FINANGLE (string);
   if (GIVENP (start))
     {
-      ASSERT (start, EXACTP (start), 3);
+      ASSERT_EXACT (3, start);
       cstart = C_INT (start);
-      if (clen < cstart)
+      if (RLEN (string) < cstart)
         ERROR ("String start index out of range: ~A", start);
     }
-  return NUM_INT ((clen == cstart)
-                  ? 0
-                  /* Somewhere around PostgreSQL 8.1 the first arg type
-                     changed from ‘const unsigned char *’ to ‘const char *’.
-                     This ‘void *’ papers over that wart.  */
-                  : PQmblen ((void *) (ROZT (string) + cstart), cenc));
+  rv = (RLEN (string) == cstart
+        ? 0
+        /* Somewhere around PostgreSQL 8.1 the first arg type
+           changed from ‘const unsigned char *’ to ‘const char *’.
+           This ‘void *’ papers over that wart.  */
+        : PQmblen ((void *) (RS (string) + cstart), cenc));
+  UNFINANGLE (string);
+  return NUM_INT (rv);
 #undef FUNC_NAME
 }
 
@@ -2996,12 +3193,16 @@ Return @code{#t} if successful, @code{#f} otherwise.  */)
 {
 #define FUNC_NAME s_pg_set_client_encoding_x
   PGconn *dbconn;
+  range_t cencoding;
+  int rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, encoding);
-  ROZT_X (encoding);
 
-  return BOOLEAN (! PQsetClientEncoding (dbconn, ROZT (encoding)));
+  FINANGLE (encoding);
+  rv = PQsetClientEncoding (dbconn, RS (encoding));
+  UNFINANGLE (encoding);
+  return BOOLEAN (! rv);
 #undef FUNC_NAME
 }
 
@@ -3057,12 +3258,16 @@ message is retrievable with @code{pg-error-message}.  */)
 {
 #define FUNC_NAME s_pg_send_query
   PGconn *dbconn;
+  range_t cquery;
+  int rv;
 
   VALIDATE_CONNECTION_UNBOX_DBCONN (1, conn, dbconn);
   ASSERT_STRING (2, query);
-  ROZT_X (query);
 
-  return BOOLEAN (PQsendQuery (dbconn, ROZT (query)));
+  FINANGLE (query);
+  rv = PQsendQuery (dbconn, RS (query));
+  UNFINANGLE (query);
+  return BOOLEAN (rv);
 #undef FUNC_NAME
 }
 
@@ -3075,6 +3280,7 @@ parameterized string, and @var{parms} is a parameter-vector.  */)
 {
 #define FUNC_NAME s_pg_send_query_params
   PGconn *dbconn;
+  range_t cquery;
   struct paramspecs ps;
   int result;
 
@@ -3082,9 +3288,10 @@ parameterized string, and @var{parms} is a parameter-vector.  */)
 
   prep_paramspecs (FUNC_NAME, &ps, parms);
   NOINTS ();
-  result = PQsendQueryParams (dbconn, ROZT (query), ps.len,
+  result = PQsendQueryParams (dbconn, RS (query), ps.len,
                               ps.types, ps.values, ps.lengths, ps.formats,
                               RESFMT_TEXT);
+  UNFINANGLE (query);
   INTSOK ();
   drop_paramspecs (&ps);
   return BOOLEAN (result);
@@ -3100,6 +3307,7 @@ Also, return @code{#t} if successful.  */)
 {
 #define FUNC_NAME s_pg_send_query_prepared
   PGconn *dbconn;
+  range_t cstname;
   struct paramspecs ps;
   int result;
 
@@ -3107,9 +3315,10 @@ Also, return @code{#t} if successful.  */)
 
   prep_paramspecs (FUNC_NAME, &ps, parms);
   NOINTS ();
-  result = PQsendQueryPrepared (dbconn, ROZT (stname), ps.len,
+  result = PQsendQueryPrepared (dbconn, RS (stname), ps.len,
                                 ps.values, ps.lengths, ps.formats,
                                 RESFMT_TEXT);
+  UNFINANGLE (stname);
   INTSOK ();
   drop_paramspecs (&ps);
   return BOOLEAN (result);
@@ -3211,26 +3420,6 @@ data was sent (only possible for a non-blocking connection).  */)
 #undef FUNC_NAME
 }
 
-
-/*
- * installation features
- */
-
-#define SIMPLE_SYMBOL(s)                        \
-  SCM_SYMBOL (pg_sym_ ## s, # s)
-
-#define SYM(s)  (pg_sym_ ## s)
-
-SIMPLE_SYMBOL (PQPROTOCOLVERSION);
-SIMPLE_SYMBOL (PQRESULTERRORMESSAGE);
-SIMPLE_SYMBOL (PQPASS);
-SIMPLE_SYMBOL (PQBACKENDPID);
-SIMPLE_SYMBOL (PQOIDVALUE);
-SIMPLE_SYMBOL (PQBINARYTUPLES);
-SIMPLE_SYMBOL (PQFMOD);
-SIMPLE_SYMBOL (PQSETNONBLOCKING);
-SIMPLE_SYMBOL (PQISNONBLOCKING);
-
 
 /*
  * init
@@ -3240,9 +3429,6 @@ static
 void
 init_module (void)
 {
-#define DEFSMOB(tagvar,name,m,f,p)                              \
-  tagvar = scm_make_smob_type_mfpe (name, 0, m, f, p, NULL)
-
   DEFSMOB (pg_conn_tag, xc_name,
            xc_mark,
            xc_free,
@@ -3257,7 +3443,6 @@ init_module (void)
            NULL,
            sepo_free,
            sepo_display);
-#undef DEFSMOB
 
 #include "libpq.x"
 
@@ -3300,17 +3485,6 @@ init_module (void)
   scm_set_port_truncate      (lobp_tag, NULL);
   scm_set_port_input_waiting (lobp_tag, lob_input_waiting_p);
 
-  goodies
-    = KLIST (SYM (PQPROTOCOLVERSION),
-             SYM (PQRESULTERRORMESSAGE),
-             SYM (PQPASS),
-             SYM (PQBACKENDPID),
-             SYM (PQOIDVALUE),
-             SYM (PQBINARYTUPLES),
-             SYM (PQFMOD),
-             SYM (PQSETNONBLOCKING),
-             SYM (PQISNONBLOCKING));
-
 #undef KLIST
 }
 
