@@ -46,6 +46,7 @@
             sql-pre
             sql-pre?
             sql-unpre
+            sql-quote-auto-E?
             sql-quote
             idquote
             make-comma-separated-tree
@@ -191,6 +192,31 @@
   (set! (--preformatted string) #f)
   string)
 
+;; When non-@code{#f}, @code{sql-quote} will prefix a @samp{E}
+;; if it detects any @samp{\} (backslash) characters in its arg.
+;; For example:
+;;
+;; @example
+;; (define (e.g.)
+;;   (display (sql-quote "a\\b")))
+;;
+;; (fluid-ref sql-quote-auto-E?)
+;; @result{} #f
+;;
+;; (e.g.)
+;; @print{} 'a\b'
+;;
+;; (with-fluids ((sql-quote-auto-E? #t))
+;;   (e.g.))
+;; @print{} E'a\b'
+;; @end example
+;;
+;; If there are no backslash characters, this has no effect.
+;;
+;;-category: fluid
+;;
+(define sql-quote-auto-E? (make-fluid))
+
 ;; Return a new string made by doubling each single-quote in string
 ;; @var{s}, and prefixing and suffixing with single-quote.
 ;; The returned string is marked by @code{sql-pre}.  For example:
@@ -213,23 +239,30 @@
 (define (sql-quote s)
   (or (string? s) (error "not a string:" s))
   (let* ((olen (string-length s))
-         (len (+ 2 olen))
+         (E? (and (fluid-ref sql-quote-auto-E?)
+                  (string-index s #\\)))
+         (left-pos (if E? 1 0))
+         (len (+ left-pos 2 olen))
          (cuts (let loop ((stop olen) (acc (list olen)))
+                 (define (with x)
+                   (acons x (car acc) (cdr acc)))
                  (cond ((string-rindex s #\' 0 stop)
                         => (lambda (hit)
                              (set! len (1+ len))
-                             (loop hit (cons hit (cons hit acc)))))
-                       (else (cons 0 acc)))))
+                             (loop hit (cons hit (with hit)))))
+                       (else (with 0)))))
          (rv (make-string len)))
-    (string-set! rv 0 #\')
+    (and E? (string-set! rv 0 #\E))
+    (string-set! rv left-pos #\')
     (string-set! rv (1- len) #\')
-    (let loop ((put 1) (ls cuts))
+    (let loop ((put (1+ left-pos)) (ls cuts))
       (if (null? ls)
           (sql-pre rv)                  ; rv
-          (let* ((one (car ls))
-                 (two (cadr ls))
+          (let* ((pair (car ls))
+                 (one (car pair))
+                 (two (cdr pair))
                  (end (+ put (- two one)))
-                 (tail (cddr ls)))
+                 (tail (cdr ls)))
             (substring-move! s one two rv put)
             (or (null? tail)
                 (string-set! rv end #\'))
@@ -251,6 +284,9 @@
 ;; (try 'ab.cd)      @print{} "ab"."cd"
 ;; (try 'abcd[xyz])  @print{} "abcd"[xyz]
 ;; (try 'ab.cd[xyz]) @print{} "ab"."cd"[xyz]
+;;
+;; ;; Special case: only * after dot.
+;; (try 'ab.*)       @print{} "ab".*
 ;; @end lisp
 ;;
 ;; Note that PostgreSQL case-folding for non-quoted identifiers
@@ -276,11 +312,14 @@
       ;; Fast path; no complications.
       ((not (or ra dot))
        (object->string s))
-      ;; Just dot (ab.cd => "ab"."cd").
+      ;; Just dot (ab.cd => "ab"."cd", but ab.* => "ab".*).
       ((and dot (not ra))
        (fs "~S.~S"
            (substring s 0 dot)
-           (substring s (1+ dot))))
+           (let ((after (substring s (1+ dot))))
+             (if (string=? "*" after)
+                 '*
+                 after))))
       ;; Just array (abcd[xyz] => "abcd"[xyz]).
       ((and ra (not dot))
        (fs "~S~A"
