@@ -177,7 +177,6 @@
     (lambda ()
       (let ((v (pg-protocol-version *C*)))
         (and (number? v)
-             (not (= 0 v))
              (< 1 v))))))
 
 (define test:tracing-traced-connection
@@ -267,7 +266,7 @@
 (define test:parameter-status
   (add-test #t
     (lambda ()
-      (or (not (memq 'PQPROTOCOLVERSION (pg-guile-pg-loaded)))
+      (or (= 2 (pg-protocol-version *C*))
           (and-map (lambda (k)
                      (let ((v (pg-parameter-status *C* k)))
                        (format #t "INFO: parameter ~S => ~S\n" k v)
@@ -302,7 +301,7 @@
   (add-test #:default
     (lambda ()
       (and (pg-set-error-verbosity *C* #:terse)
-           (eq? (if (memq 'PQPROTOCOLVERSION (pg-guile-pg-loaded))
+           (eq? (if (< 2 (pg-protocol-version *C*))
                     #:terse
                     #:default)
                 (pg-set-error-verbosity *C* #:default))
@@ -504,10 +503,10 @@
              (eq? (pg-fnumber res "col1") 0)
              (eq? (pg-fnumber res "col2") 1)
              (eq? (pg-fnumber res "invalid_column_name") -1)
-             (or (not (memq 'PQPROTOCOLVERSION (pg-guile-pg-loaded)))
+             (or (= 2 (pg-protocol-version *C*))
                  (and (integer? (pg-ftable res 0))
                       (integer? (pg-ftablecol res 0))
-                      (= 0 (pg-fformat res 0)))))))))
+                      (zero? (pg-fformat res 0)))))))))
 
 (define test:getvalue
   (add-test #t
@@ -619,7 +618,7 @@
 (define test:get-copy-data
   (add-test #t
     (lambda ()
-      (or (not (memq 'PQPROTOCOLVERSION (pg-guile-pg-loaded)))
+      (or (= 2 (pg-protocol-version *C*))
           (and (command-ok?
                 (cexec "SELECT * INTO t1 FROM test WHERE col1 = 1"))
                (let ((res (cexec "COPY t1 TO STDOUT"))
@@ -746,12 +745,12 @@
        (command-ok? (cexec "CREATE TABLE async (a numeric (20, 10))"))
        ((ok? 'PGRES_COPY_IN) (cexec "COPY async FROM STDIN"))
        (begin
-         (and (memq 'PQPROTOCOLVERSION (pg-guile-pg-loaded))
+         (and (< 2 (pg-protocol-version *C*))
               ;; Test fails for PostgreSQL 7.4.12 at ‘pg-endcopy’ if omitted.
               ;; This needs to be done prior to ‘pg-putline’, as well.
               (pg-set-nonblocking! *C* #f))
          (do ((i 4224 (1- i)))
-             ((= i 0))
+             ((zero? i))
            (pg-putline *C* (format #f "~A.~A\n" i i)))
          (pg-putline *C* "\\.\n")
          (pg-endcopy *C*))
@@ -790,7 +789,7 @@
        (not (pg-is-busy? *C*))
        ;; Start a transaction that's easy to cancel.
        ;; [Insert cynical comparison to inept manglement/politicos here.]
-       (pg-send-query *C* "BEGIN; UPDATE async SET a = ! a * sqrt (a);")
+       (pg-send-query *C* "BEGIN; UPDATE async SET a = a * sqrt (a);")
        ;; Dispatch goes through.
        (pg-request-cancel *C*)
        ;; Result handling, however, may finish, or may signal error.
